# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from datetime import datetime

from sqlalchemy.orm import Session
from sqlalchemy.sql import false

from . import models, schemas


def get_workflow(db: Session, workflow_id: int):
    return (
        db.query(models.Workflows)
        .filter(models.Workflows.id == workflow_id)
        .first()
    )


def get_workflow_by_name(db: Session, name: str):
    return (
        db.query(models.Workflows)
        .filter(
            models.Workflows.name == name,
            models.Workflows.delete_flag == false(),
        )
        .first()
    )


def create_workflow(
    db: Session, workflow: schemas.WorkflowCreate
) -> models.Workflows:
    db_workflow = models.Workflows(
        name=workflow.name,
        desc=workflow.desc,
        params=workflow.params,
        on=workflow.on,
        jobs=workflow.jobs,
        valid_start=datetime.now(),
        valid_end=datetime(2999, 12, 31),
    )
    db.add(db_workflow)
    db.commit()
    db.refresh(db_workflow)
    return db_workflow


def list_workflows(db: Session, skip: int = 0, limit: int = 1000):
    return (
        db.query(models.Workflows)
        .filter(models.Workflows.delete_flag == false())
        .offset(skip)
        .limit(limit)
        .all()
    )


def search_workflow(db: Session, search_text: str):
    if len(search_text) > 1:
        if not (search_text := search_text.strip().lower()):
            return []

        results = []
        for workflow in list_workflows(db=db):
            text: str = f"{workflow.name} {workflow.desc or ''}".lower()
            if search_text in text:
                results.append(workflow)
        return results
    return list_workflows(db=db)



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from contextlib import asynccontextmanager
from typing import Annotated, Optional

from fastapi import APIRouter, Depends, FastAPI, Header, HTTPException, Request
from fastapi import status as st
from fastapi.templating import Jinja2Templates
from sqlalchemy.orm import Session

from ...db import engine
from ...deps import get_db, get_templates
from ...utils import get_logger
from . import crud, models
from .schemas import Workflow, WorkflowCreate, Workflows

logger = get_logger("ddeutil.observe")


@asynccontextmanager
async def lifespan(_: FastAPI):
    """Lifespan for create workflow tables on target database."""
    models.Base.metadata.create_all(bind=engine)
    yield


workflow = APIRouter(prefix="/workflow", tags=["workflow"], lifespan=lifespan)


@workflow.get("/")
def read_workflows(
    request: Request,
    db: Session = Depends(get_db),
    templates=Depends(get_templates),
):
    """Return all workflows."""
    workflows: list[Workflow] = Workflows.validate_python(
        crud.list_workflows(db)
    )
    return templates.TemplateResponse(
        request=request,
        name="workflow/index.html",
        context={
            "workflows": workflows,
            "search_text": "",
        },
    )


@workflow.post("/", response_model=Workflow)
def create_workflow(wf: WorkflowCreate, db: Session = Depends(get_db)):
    db_workflow = crud.get_workflow_by_name(db, name=wf.name)
    if db_workflow:
        raise HTTPException(
            status_code=st.HTTP_302_FOUND,
            detail="Workflow already registered to observe database.",
        )
    return crud.create_workflow(db=db, workflow=wf)


@workflow.get("/search")
def search_workflows(
    request: Request,
    search_text: str,
    hx_request: Annotated[Optional[str], Header(...)] = None,
    db: Session = Depends(get_db),
    templates: Jinja2Templates = Depends(get_templates),
):
    workflows: list[Workflow] = Workflows.validate_python(
        crud.search_workflow(db=db, search_text=search_text)
    )
    if hx_request:
        return templates.TemplateResponse(
            "workflow/partials/search_results.html",
            {"request": request, "workflows": workflows},
        )
    return templates.TemplateResponse(
        request=request,
        name="workflow/index.html",
        context={
            "workflows": workflows,
            "search_text": search_text,
        },
    )



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from dotenv import load_dotenv
from fastapi import Depends, FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from sqlalchemy.orm import Session

from .__about__ import __version__
from .deps import get_db, get_templates
from .routes import log, workflow
from .utils import get_logger

load_dotenv()
logger = get_logger("ddeutil.observe")

app = FastAPI(
    titile="Observe Web",
    version=__version__,
)
app.add_middleware(
    CORSMiddleware,
    allow_origins=[
        "http://localhost",
        "http://localhost:8080",
    ],
    allow_methods=["*"],
    allow_headers=["*"],
)
app.include_router(workflow)
app.include_router(log)
app.mount("/static", StaticFiles(directory="static"), name="static")


@app.get("/", response_class=HTMLResponse)
async def index(
    request: Request,
    templates: Jinja2Templates = Depends(get_templates),
    db: Session = Depends(get_db),
):
    return templates.TemplateResponse(
        request=request, name="home/index.html", context={}
    )



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from sqlalchemy import (
    JSON,
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    String,
)
from sqlalchemy.orm import relationship

from ...db import Base


class Workflows(Base):
    __tablename__ = "workflows"

    id = Column(Integer, primary_key=True, index=True)
    name = Column(String, index=True)
    desc = Column(String)
    params = Column(JSON)
    on = Column(JSON)
    jobs = Column(JSON)
    delete_flag = Column(Boolean, default=False)
    valid_start = Column(DateTime)
    valid_end = Column(DateTime)

    releases = relationship("WorkflowReleases", back_populates="workflow")


class WorkflowReleases(Base):
    __tablename__ = "workflow_releases"

    id = Column(Integer, primary_key=True, index=True)
    release = Column(DateTime, index=True)
    workflow_id = Column(Integer, ForeignKey("workflows.id"))

    workflow = relationship("Workflows", back_populates="releases")
    logs = relationship("WorkflowLogs", back_populates="release")



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import pathlib

from fastapi import Request
from fastapi.templating import Jinja2Templates
from jinja2 import ChoiceLoader, FileSystemLoader

from .db import SessionLocal


def get_templates(request: Request) -> Jinja2Templates:
    """Dynamic multi-templating Jinja2 loader that support templates inside
    APIRouter.
    """
    choices: list[FileSystemLoader] = [FileSystemLoader("./templates")]
    if request.url.path != "/":
        route: str = request.url.path.strip("/").split("/")[0]
        route_path: pathlib.Path = (
            pathlib.Path(__file__).parent / f"routes/{route}/templates"
        )
        if route_path.exists():
            choices.insert(0, FileSystemLoader(route_path))

    return Jinja2Templates(
        directory="templates",
        loader=ChoiceLoader(choices),
    )


def get_db() -> SessionLocal:
    """Return the database local session instance."""
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import os
from typing import Any

from sqlalchemy import create_engine, event
from sqlalchemy.engine import Engine
from sqlalchemy.orm import declarative_base, sessionmaker


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, _):
    """Read more:"""
    cursor = dbapi_connection.cursor()
    settings: dict[str, Any] = {
        "journal_mode": "WAL",
        "foreign_keys": "ON",
        "page_size": 4096,
        "cache_size": 10000,
        # "locking_mode": 'EXCLUSIVE',
        "synchronous": "NORMAL",
    }
    for k, v in settings.items():
        cursor.execute(f"PRAGMA {k} = {v};")
    cursor.close()


SQLALCHEMY_DATABASE_URL: str = os.getenv(
    "OBSERVE_SQLALCHEMY_DATABASE_URL", "sqlite:///./observe.db"
)

engine = create_engine(
    SQLALCHEMY_DATABASE_URL,
    connect_args={"check_same_thread": False},
)
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)

Base = declarative_base()



"""
To avoid confusion between the SQLAlchemy models and the Pydantic models, we
will have the file models.py with the SQLAlchemy models, and the file schemas.py
with the Pydantic models.

These Pydantic models define more or less a "schema" (a valid data shape).

So this will help us avoiding confusion while using both.

Read more: https://fastapi.tiangolo.com/tutorial/sql-databases/?h=database
"""

from __future__ import annotations

from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, TypeAdapter


class WorkflowBase(BaseModel):
    name: str
    desc: Optional[str] = None
    params: dict[str, Any]
    on: list[dict[str, Any]]
    jobs: dict[str, Any]


class WorkflowCreate(WorkflowBase):
    pass


class Workflow(WorkflowBase):
    model_config = ConfigDict(from_attributes=True)

    id: int


Workflows = TypeAdapter(list[Workflow])


class ReleaseBase(BaseModel):
    release: datetime


class Release(ReleaseBase):
    model_config = ConfigDict(from_attributes=True)

    id: int



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from sqlalchemy import JSON, Column, DateTime, ForeignKey, String
from sqlalchemy.orm import relationship

from ...db import Base


class WorkflowLogs(Base):
    __tablename__ = "workflow_logs"

    run_id = Column(String, primary_key=True, index=True)
    log = Column(JSON)
    release_id = Column(DateTime, ForeignKey("workflow_releases.id"))

    release = relationship("WorkflowReleases", back_populates="logs")



from ddeutil.observe.routes.workflow.schemas import WorkflowCreate


def test_workflow_schema():
    # NOTE: This data return from ddeutil-workflow api.
    value = {
        "name": "wf-scheduling",
        "params": {"asat-dt": {"type": "datetime"}},
        "on": [
            {"cronjob": "*/3 * * * *", "timezone": "Asia/Bangkok"},
            {"cronjob": "* * * * *", "timezone": "Asia/Bangkok"},
        ],
        "jobs": {
            "condition-job": {
                "id": "condition-job",
                "stages": [
                    {"name": "Empty stage"},
                    {
                        "name": "Call-out",
                        "echo": "Hello ${{ params.asat-dt | fmt('%Y-%m-%d') }}",
                    },
                ],
            }
        },
    }
    rs = WorkflowCreate.model_validate(value)
    print(rs)



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from contextlib import asynccontextmanager
from typing import Annotated, Optional

from fastapi import APIRouter, Depends, FastAPI, Header, Request

from ...db import engine
from ...deps import get_templates
from . import models


@asynccontextmanager
async def lifespan(_: FastAPI):
    """Lifespan for create log tables on target database."""
    models.Base.metadata.create_all(bind=engine)
    yield


log = APIRouter(prefix="/log", tags=["log"], lifespan=lifespan)


@log.get("/")
def read_logs(
    request: Request,
    hx_request: Annotated[Optional[str], Header(...)] = None,
    templates=Depends(get_templates),
):
    """Return all workflows."""
    if hx_request:
        return templates.TemplateResponse(
            "log/partials/show_add_author_form.html", {"request": request}
        )
    return templates.TemplateResponse(request=request, name="log/index.html")
